__all__ = ('DiscordEntity',)

from scarletio import RichAttributeErrorBaseType, include

from .discord_entity_meta import DiscordEntityMeta, ENTITY_ID_PLACEHOLDER


id_to_datetime = include('id_to_datetime')


class DiscordEntity(RichAttributeErrorBaseType, metaclass = DiscordEntityMeta):
    """
    Base type for Discord entities.
    
    Notes
    -----
    Inherit it with passing `immortal = True` to make the sub-types weakreferable.
    """
    id = ENTITY_ID_PLACEHOLDER
    
    __slots__ = ()
    
    @property
    def created_at(self):
        """
        When the entity was created.
        
        Returns
        -------
        created_at : `DateTime`
        """
        return id_to_datetime(self.id)
    
    
    def __hash__(self):
        """Returns the has value of the entity, what equals to it's id."""
        return self.id
    
    
    def __gt__(self, other):
        """Whether this entity's id is greater than the other's."""
        if type(self) is type(other):
            return self.id > other.id
        
        return NotImplemented
    
    
    def __ge__(self, other):
        """Whether this entity's id is greater or equal than the other's."""
        if type(self) is type(other):
            return self.id >= other.id
        
        return NotImplemented
    
    
    def __eq__(self, other):
        """Whether this entity's id is equal as the other's."""
        if type(self) is type(other):
            return self.id == other.id
        
        return NotImplemented
    
    
    def __ne__(self, other):
        """Whether this entity's id is not equal as the other's."""
        if type(self) is type(other):
            return self.id != other.id
        
        return NotImplemented
    
    
    def __le__(self, other):
        """Whether this entity's id is less or equal than the other's."""
        if type(self) is type(other):
            return self.id <= other.id
        
        return NotImplemented
    
    
    def __lt__(self, other):
        """Whether this entity's id is less than the other's."""
        if type(self) is type(other):
            return self.id < other.id
        
        return NotImplemented



__all__ = ('PLUGIN_AUTO_RELOADER_MANAGER', 'PluginAutoReloaderManager', )

from os import listdir as list_directory
from os.path import dirname as get_parent_directory_path, isdir as is_directory, isfile as is_file, join as join_paths

from scarletio import LOOP_TIME, RichAttributeErrorBaseType, Task, write_exception_async

from ...discord import KOKORO

from ..plugin_loader import PluginError, frame_filter, get_plugin, register_plugin, reload_plugin
from ..plugin_loader.constants import IGNORED_DIRECTORY_NAMES

from .compatibility import AUTO_RELOAD_SUPPORTED, INotify
from .constants import (
    AUTO_RELOAD_DELAY, WATCH_MASK_CREATE, WATCH_MASK_DELETE, WATCH_MASK_DELETE_SELF, WATCH_MASK_GENERAL,
    WATCH_MASK_MOVE_FROM, WATCH_MASK_MOVE_SELF, WATCH_MASK_MOVE_TO, WATCH_MASK_UPDATE
)
from .helpers import _iter_plugin_root_paths, _iter_sub_directories


class PluginAutoReloaderManager(RichAttributeErrorBaseType):
    """
    Plugin auto reloader manager.
    
    Attributes
    ----------
    _directory_watch_identifier_pairs : `set` of `tuple` (`str`, `int`)
        Directory path - watcher identifier pairs.
    _inotify : `None`, `inotify_simple.Inotify`
        Inotify instance used by the auto reloader.
    _plugin_root_paths : `set<str>`
        Paths to plugin roots.
    _reload_callback_handle : `None`, ``TimerHandle``
        Handle that delays reload.
    _updated_plugins : `set` of ``Plugin``
        Actioned plugins.
    _updated_plugins_at : `float`
        When the last actioned plugin was received in loop time.
    _reload_task_active : `bool`
        Whether reload task is active.
    """
    __slots__ = (
        '_directory_watch_identifier_pairs', '_inotify', '_plugin_root_paths', '_reload_callback_handle',
        '_reload_task_active', '_updated_plugins', '_updated_plugins_at'
    )
    
    _instance = None
    
    def __new__(cls):
        """
        Creates a new plugin reloader manager instance.
        If the type is already instanced, returns the existing instance instead.
        """
        self = cls._instance
        if self is not None:
            return self
        
        self = object.__new__(cls)
        self._directory_watch_identifier_pairs = set()
        self._inotify = None
        self._plugin_root_paths = set()
        self._reload_callback_handle = None
        self._reload_task_active = False
        self._updated_plugins = set()
        self._updated_plugins_at = -AUTO_RELOAD_DELAY
        
        cls._instance = self
        return self
    
    
    def start(self):
        """
        Starts the auto reload manager.
        
        Returns
        -------
        success : `bool`
        """
        inotify = self._inotify
        if inotify is not None:
            return False
        
        if not AUTO_RELOAD_SUPPORTED:
            return False
        
        inotify = INotify(nonblocking = True)
        self._inotify = inotify
        KOKORO.add_reader(inotify.fd, self._reader_callback)
        
        self.update()
        return True
    
    
    def update(self):
        """
        Updates the auto reloaded directories.
        
        Returns
        -------
        success : `bool`
        """
        inotify = self._inotify
        if inotify is None:
            return False
        
        plugin_root_paths = self._plugin_root_paths
        plugin_root_paths.clear()
        plugin_root_paths.update(_iter_plugin_root_paths())
        
        old_directories = {item[0] for item in self._directory_watch_identifier_pairs}
        new_directories = {*_iter_sub_directories(plugin_root_paths)}
        
        success = True
        
        for directory in new_directories - old_directories:
            if not self._add_directory(directory):
                success = False
        
        for directory in old_directories - new_directories:
            if not self._remove_directory(directory):
                success = False
        
        return success
    
    
    def stop(self):
        """
        Stops the auto reloader.
        
        Returns
        -------
        success : `bool`
        """
        inotify = self._inotify
        if (inotify is None):
            return False
        
        self._inotify = None
        KOKORO.remove_reader(inotify.fd)
        inotify.close()
        
        reload_callback_handle = self._reload_callback_handle
        if (reload_callback_handle is not None):
            self._reload_callback_handle = None
            reload_callback_handle.cancel()
    
    
    # Operations
    
    def _get_directory_by_watch_identifier(self, watch_identifier):
        """
        Gets directory by watch identifier.
        
        Parameters
        ----------
        watch_identifier : `int`
            Watch identifier.
        
        Returns
        -------
        directory : `None`, `str`
            The matched directory. `None` if not found.
        """
        for iterated_directory, iterated_watch_identifier in self._directory_watch_identifier_pairs:
            if iterated_watch_identifier == watch_identifier:
                return iterated_directory
    
    
    def _get_watch_identifier_by_directory(self, directory):
        """
        Gets watch identifier by directory.
        
        Parameters
        ----------
        directory : `str`
            Directory path.
        
        Returns
        -------
        watch_identifier : `int`
            The matched watch identifier. `None` if not found.
        """
        for iterated_directory, iterated_watch_identifier in self._directory_watch_identifier_pairs:
            if iterated_directory == directory:
                return iterated_watch_identifier
    
    
    def _add_directory(self, directory):
        """
        Adds a new directory to watch.
        
        Parameters
        ----------
        directory : `str`
            Directory path.
        
        Returns
        -------
        success : `bool`
        """
        inotify = self._inotify
        if inotify is None:
            return False
        
        if (self._get_watch_identifier_by_directory(directory) is not None):
            return False
        
        watch_identifier = inotify.add_watch(directory, WATCH_MASK_GENERAL)
        self._directory_watch_identifier_pairs.add((directory, watch_identifier))
        return True
    
    
    def _remove_directory(self, directory):
        """
        Removes the directory from the watched ones.
        
        Parameters
        ----------
        directory : `str`
            Directory path.
        
        Returns
        -------
        success : `bool`
        """
        inotify = self._inotify
        if inotify is None:
            return False
        
        watch_identifier = self._get_watch_identifier_by_directory(directory)
        if (watch_identifier is None):
            return False
        
        try:
            inotify.rm_watch(watch_identifier)
        except OSError:
            return False
        
        self._directory_watch_identifier_pairs.discard((directory, watch_identifier))
        return True
    
    
    def _reader_callback(self):
        """
        Callback added to `inotify_simple.Inotify` read events.
        
        Reads from the file descriptors and calls ``._handle_event``.
        """
        inotify = self._inotify
        if inotify is None:
            # Should not happen
            return
        
        try:
            events = inotify.read(timeout = 0.0)
        except BlockingIOError:
            return
        
        for event in events:
            self._handle_event(event)
    
    
    def _handle_event(self, event):
        """
        Handles a received inotify event.
        
        Parameters
        ----------
        event : `inotify_simple.Event`
            Inotify event.
        """
        # Get directory
        directory = self._get_directory_by_watch_identifier(event.wd)
        if directory is None:
            # Unknown directory?
            return
        
        # If the name is empty (directory case) we dont want to join it to the directory.
        name = event.name
        if name:
            path = join_paths(directory, name)
        else:
            path = directory
        
        if is_directory(path) and (event.name not in IGNORED_DIRECTORY_NAMES):
            self._handle_directory_event(event, path)
            return
        
        if is_file(path):
            # we only want py files
            if not path.endswith('.py'):
                return
            
            self._handle_file_event(event, path)
            return
        
        mask = event.mask
        
        # If a file is deleted the above two wont trigger.
        if mask & WATCH_MASK_DELETE or mask & WATCH_MASK_MOVE_FROM:
            if path.endswith('.py'):
                self._handle_file_or_plugin_directory_update_or_delete(path)
            return
        
        # If a directory is deleted the top 3 wont trigger (they might, but we ignore them)
        if mask & WATCH_MASK_DELETE_SELF or mask & WATCH_MASK_MOVE_SELF:
            if (self._get_watch_identifier_by_directory(path) is not None):
                self._remove_directory(path)
                self._handle_file_or_plugin_directory_update_or_delete(path)
            
            return
    
    
    def _handle_directory_event(self, event, directory_path):
        """
        Handles a received inotify event for a directory.
        
        Parameters
        ----------
        event : `inotify_simple.Event`
            Inotify event.
        directory_path : `str`
            Path to the directory.
        """
        mask = event.mask
        if mask & (WATCH_MASK_DELETE_SELF | WATCH_MASK_MOVE_FROM):
            self._remove_directory(directory_path)
            return
        
        if mask & WATCH_MASK_CREATE:
            self._handle_directory_create(directory_path)
            return
        
        if mask & WATCH_MASK_MOVE_TO:
            self._handle_directory_move_to(directory_path)
            return
    
    
    def _handle_directory_create(self, directory_path):
        """
        Handles a received directory create inotify event.
        
        Parameters
        ----------
        directory_path : `str`
            Path to the directory.
        """
        self._add_directory(directory_path)
        self._handle_directory_create_initial_entities(directory_path)
    
    
    def _handle_directory_move_to(self, directory_path):
        """
        handles a directory move-to inotify event.
        
        Parameters
        ----------
        directory_path : `str`
            Path to the directory.
        """
        self._handle_directory_create(directory_path)
        
        if get_parent_directory_path(directory_path) in self._plugin_root_paths:
            init_file_path = join_paths(directory_path, '__init__.py')
            if is_file(init_file_path):
                register_plugin(directory_path)
                self._handle_file_or_plugin_directory_update_or_delete(init_file_path)
    
    
    def _handle_directory_create_initial_entities(self, directory_path):
        """
        Called after a directory is created to scan it.
        
        Parameters
        ----------
        directory_path : `str`
            Path to the directory.
        """
        for name in list_directory(directory_path):
            path = join_paths(directory_path, name)
            if is_file(path):
                self._handle_file_create(name, path)
                continue
            
            if is_directory(path) and (name not in IGNORED_DIRECTORY_NAMES):
                self._handle_directory_create(path)
                continue
    
    
    def _handle_file_event(self, event, file_path):
        """
        Handles a received inotify event for a file.
        
        Parameters
        ----------
        event : `inotify_simple.Event`
            Inotify event.
        file_path : `str`
            Path to the file.
        """
        mask = event.mask
        if mask & WATCH_MASK_CREATE:
            self._handle_file_create(event.name, file_path)
            return
        
        if mask & WATCH_MASK_MOVE_TO:
            self._handle_file_move_to_event(event.name, file_path)
            return
        
        if mask & WATCH_MASK_UPDATE:
            self._handle_file_or_plugin_directory_update_or_delete(file_path)
            return
    
    
    def _handle_file_create(self, name, file_path):
        """
        Handles a received file create inotify event.
        
        Parameters
        ----------
        name : `str`
            The file's name.
        file_path : `str`
            Path to the file.
        
        Returns
        -------
        is_registered : `bool`
            Whether the file was registered as a plugin.
        """
        # If we register an init.py file in directory of a plugin root, we register it as a plugin
        if name == '__init__.py':
            parent_directory = get_parent_directory_path(file_path)
            if get_parent_directory_path(parent_directory) in self._plugin_root_paths:
                register_plugin(parent_directory)
                return True
        
        # If we register a file inside of a plugin directory, we again wanna register it as a plugin
        else:
            if get_parent_directory_path(file_path) in self._plugin_root_paths:
                register_plugin(file_path)
                return True
        
        return False
    
    
    def _handle_file_move_to_event(self, name, file_path):
        """
        handles a received file move-to inotify event.
        
        Parameters
        ----------
        name : `str`
            The file's name.
        file_path : `str`
            Path to the file.
        """
        if self._handle_file_create(name, file_path):
            self._handle_file_or_plugin_directory_update_or_delete(file_path)
    
    
    def _handle_file_or_plugin_directory_update_or_delete(self, file_path):
        """
        Handles a received file delete / edit inotify event.
        
        Parameters
        ----------
        file_path : `str`
            Path to the file.
        """
        # Get corresponding plugin
        plugin = get_plugin(file_path)
        if plugin is None:
            return
        
        # Set file path into updated
        self._updated_plugins.add(plugin)
        
        # Set update time
        self._updated_plugins_at = LOOP_TIME()
        
        # Start callback
        reload_callback_handle = self._reload_callback_handle
        if (reload_callback_handle is None):
            self._reload_callback_handle = KOKORO.call_after(AUTO_RELOAD_DELAY, self._reload_callback)
    
    
    def _reload_callback(self):
        """
        Invokes ``._do_reload`` if no new events were triggered since last scheduling.
        If there were events triggered then reschedules itself.
        """
        next = self._updated_plugins_at + AUTO_RELOAD_DELAY
        if next > LOOP_TIME() or self._reload_task_active:
            self._reload_callback_handle = KOKORO.call_at(AUTO_RELOAD_DELAY, self._reload_callback)
            return
        
        self._reload_callback_handle = None
        Task(KOKORO, self._do_reload())
    
    
    async def _do_reload(self):
        """
        Keeps reloading plugins till there are new plugins to reload.
        
        This method is a generator.
        """
        if self._reload_task_active:
            return
        
        self._reload_task_active = True
        try:
            while True:
                self_updated_plugins = self._updated_plugins
                if not self_updated_plugins:
                    return
                
                updated_plugins = self_updated_plugins.copy()
                self_updated_plugins.clear()
                
                try:
                    await reload_plugin(updated_plugins)
                except GeneratorExit:
                    raise
                
                except PluginError as err:
                    await write_exception_async(err, filter = frame_filter)
                    
                except BaseException as err:
                    await write_exception_async(err)
                    return
        
        finally:
            self._reload_task_active = False


PLUGIN_AUTO_RELOADER_MANAGER = PluginAutoReloaderManager()



__all__ = ()

from .base import DiscordGatewayBase


class DiscordGatewayVoiceBase(DiscordGatewayBase):
    """
    Base gateway for voice clients.
    """
    __slots__ = ()
    
    async def set_speaking(self, speaking):
        """
        Sends a `speaking` packet.
        
        This method is a coroutine.
        
        Parameters
        ----------
        speaking : `bool`
            Whether the voice voice client should show up as speaking and be able to send voice data or not.
        """
        pass
    
    
    async def run(self, waiter = None):
        """
        Keeps the gateway receiving message and processing it. If the gateway needs to be reconnected, reconnects
        itself. If connecting cannot succeed, because there is no internet returns `False`. If the `.voice_client` is
        disconnected, then returns `False`.
        
        This method is a coroutine.
        
        Parameters
        -----------
        waiter : `None | Future<bool>` = `None`, Optional
            A waiter future what is set, when the gateway finished connecting and started polling events.
            Its result is also set in case the task is finished before connection.
        
        Returns
        -------
        outcome : `bool<False>`
            Always `False`.
        """
        if (waiter is not None):
            waiter.set_result_if_pending(False)
            waiter = None
        
        return False



__all__ = ('CONVERSION_VOICE_ATTACHMENT',)

from scarletio.web_common import FormData

from .....env import API_VERSION

from ....builder.constants import CONVERSION_KIND_FIELD
from ....builder.conversion import Conversion

from ...message import MessageFlag
from ...voice_attachment import VoiceAttachment


MESSAGE_FLAG_VOICE_MESSAGE = MessageFlag().update_by_keys(voice_message = True)


def _preprocess_data_and_check_empty(data, voice_attachment):
    """
    Preprocesses the data and checks whether it is empty.
    
    Parameters
    ----------
    data : ``dict<str, object> | FormData>``
        The data to preprocess.
    
    Returns
    -------
    form : ``FormData``
    
    Raises
    ------
    ValueError
        - if the given `data` is not empty.
    """
    while True:
        if isinstance(data, dict):
            if (
                ('content' in data) or
                ('embed' in data) or
                ('attachments' in data) or
                ('components' in data) or
                ('poll' in data)
            ):
                break
            
            data['attachments'] = [voice_attachment.to_data()]
            data['flags'] = data.get('flags', 0) | MESSAGE_FLAG_VOICE_MESSAGE
            
            form = FormData()
            form.add_json('payload_json', data)
        
        elif isinstance(data, FormData):
            break
        
        else:
            # Unexpected case
            break
        
        return form
    
    raise ValueError(
        f'When sending voice messages, `voice_attachment` should be the only set field.'
    )


class CONVERSION_VOICE_ATTACHMENT(Conversion):
    # Generic
    
    name = 'voice_attachment'
    name_aliases = None
    expected_types_messages = VoiceAttachment.__name__
    kind = CONVERSION_KIND_FIELD
    output_conversion = None
    
    # Setting
    
    set_merger = None
    set_type = VoiceAttachment
    set_type_processor = None
    set_identifier = None
    set_listing_identifier = None
    
    def set_validator(value):
        """
        Yields the outcome if `value` is acceptable.
        
        Parameters
        ----------
        value : ``VoiceAttachment``
            The value to check.
        
        Yields
        ------
        voice_attachment : ``None | VoiceAttachment``
        """
        # None
        if value is None:
            yield None
            return
        
        # VoiceAttachment
        if isinstance(value, VoiceAttachment):
            yield value
            return
    
    
    # Reading
    
    get_default = None
    get_processor = None
    
    
    # Serialization
    
    serializer_key = None
    serializer_optional = None
    serializer_required = None
    
    
    if API_VERSION >= 9:
        def serializer_putter(data, required, value):
            if value is None:
                return data
            
            form = _preprocess_data_and_check_empty(data, value)
            form.add_field('files[0]', value.io, file_name = value.name, content_type = 'application/octet-stream')
            return form
    
    else:
        # On lower versions voice attachments are not supported, attach them as regular attachments instead.
        def serializer_putter(data, required, value):
            if value is None:
                return data
            
            form = _preprocess_data_and_check_empty(data, value)
            form.add_field('file', value.io, file_name = value.name, content_type = 'application/octet-stream')
            return form
    
    
    # Sorting
    
    sort_priority = 10000



__all__ = ()

from scarletio import CauseGroup, copy_docs

from ...discord.events.handling_helpers import Router

from .interfaces.autocomplete import AutocompleteInterface
from .interfaces.exception_handler import ExceptionHandlerInterface
from .interfaces.nestable import NestableInterface


class InteractionCommandRouter(AutocompleteInterface, ExceptionHandlerInterface, NestableInterface, Router):
    """
    Router for returned grouped interactions.
    """
    __slots__ = ()
    
    # ---- AutocompleteInterface ----
    
    def _check_supports_auto_completion(self):
        for instance in self:
            if not isinstance(instance, AutocompleteInterface):
                raise RuntimeError(
                    f'At least 1 instance of the router does not supports auto completion; got {instance!r}.'
                )
    
    
    @copy_docs(AutocompleteInterface.autocomplete)
    def autocomplete(self, parameter_name, *parameter_names, function = ...):
        self._check_supports_auto_completion()
        return AutocompleteInterface.autocomplete(self, parameter_name, *parameter_names, function = function)
    
    
    @copy_docs(AutocompleteInterface._register_auto_completer)
    def _register_auto_completer(self, function, parameter_names):
        self._check_supports_auto_completion()

        exceptions = None
        added_auto_completers = []
        
        for instance in self:
            try:
                added_auto_completer = instance._register_auto_completer(function, parameter_names)
            except RuntimeError as exception:
                if exceptions is None:
                    exceptions = []
                
                exceptions.append(exception)
            
            else:
                added_auto_completers.append(added_auto_completer)
        
        
        if exceptions is not None:
            raise RuntimeError(
                'One or more exception occurred while registering autocompleter.'
            ) from CauseGroup(*exceptions)
        
        
        return type(self)(added_auto_completers)
    
    
    # ---- ExceptionHandlerInterface ----
    
    def _check_supports_exception_handler(self):
        for instance in self:
            if not isinstance(instance, ExceptionHandlerInterface):
                raise RuntimeError(
                    f'At least 1 instance of the router does not supports exception handlers; got {instance!r}.'
                )
    
    
    @copy_docs(ExceptionHandlerInterface.error)
    def error(self, exception_handler = None, *, first = False):
        self._check_supports_exception_handler()
        return ExceptionHandlerInterface.error(exception_handler, first = False)
    
    
    @copy_docs(ExceptionHandlerInterface._register_exception_handler)
    def _register_exception_handler(self, exception_handler, first):
        self._check_supports_exception_handler()
        
        exceptions = None
        
        for instance in self:
            try:
                instance._register_exception_handler(exception_handler, first)
            except RuntimeError as exception:
                if exceptions is None:
                    exceptions = []
                
                exceptions.append(exception)
        
        
        if exceptions is not None:
            raise RuntimeError(
                'One or more exception occurred while registering exception handler.'
            ) from CauseGroup(*exceptions)
        
        
        return exception_handler
    
    
    # ---- NestableInterface ----
    
    
    @copy_docs(NestableInterface._is_nestable)
    def _is_nestable(self):
        for instance in self:
            if (not isinstance(instance, NestableInterface)) or (not instance._is_nestable()):
                return False
        
        return True
    
    
    @copy_docs(NestableInterface._check_supports_nesting)
    def _check_supports_nesting(self):
        for instance in self:
            if (not isinstance(instance, NestableInterface)) or (not instance._is_nestable()):
                raise RuntimeError(
                    f'At least 1 instance of the router does not supports nesting; got {instance!r}.'
                )
    
    
    @copy_docs(NestableInterface.create_event)
    def create_event(self, function, *positional_parameters, **keyword_parameters):
        self._check_supports_nesting()
        
        exceptions = None
        commands = []
        
        for instance in self:
            try:
                command = instance.create_event(function, *positional_parameters, **keyword_parameters)
            except RuntimeError as exception:
                if exceptions is None:
                    exceptions = []
                
                exceptions.append(exception)
            
            else:
                commands.append(command)
        
        
        if exceptions is not None:
            raise RuntimeError(
                'One or more exception occurred while nesting.'
            ) from CauseGroup(*exceptions)
        
        
        return type(self)(commands)



__all__ = ()

from .fields import (
    put_description, put_enabled, put_steps, validate_description, validate_enabled, validate_steps
)


VERIFICATION_SCREEN_FIELD_CONVERTERS = {
    'description': (validate_description, put_description),
    'enabled': (validate_enabled, put_enabled),
    'steps': (validate_steps, put_steps),
}



__all__ = ('SKUEnhancementGuild',)

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_additional_emoji_slots, parse_additional_soundboard_sound_slots, parse_additional_sticker_slots,
    parse_features, put_additional_emoji_slots, put_additional_soundboard_sound_slots, put_additional_sticker_slots,
    put_features, validate_additional_emoji_slots, validate_additional_soundboard_sound_slots,
    validate_additional_sticker_slots, validate_features
)


class SKUEnhancementGuild(RichAttributeErrorBaseType):
    """
    Represents an applied enhancements to a guild of an SKU.
    
    Attributes
    ----------
    additional_emoji_slots : `int`
        Additional granted emoji slots by the enhancement.
    
    additional_soundboard_sound_slots : `int`
        Additionally granted soundboard sound slots by the enhancement.
    
    additional_sticker_slots : `int`
        Additionally granted sticker by the enhancement.
    
    features : ``None | tuple<GuildFeature>``
        The granted features by the enhancement.
    """
    __slots__ = ('additional_emoji_slots', 'additional_soundboard_sound_slots', 'additional_sticker_slots', 'features')
    
    def __new__(
        cls,
        *,
        additional_emoji_slots = ...,
        additional_soundboard_sound_slots = ...,
        additional_sticker_slots = ...,
        features = ...,
    ):
        """
        Creates a new user SKU enhancement guild instance from the given parameters.
        
        Attributes
        ----------
        additional_emoji_slots : `None | int`, Optional (Keyword only)
            Additional granted emoji slots by the enhancement.
        
        additional_soundboard_sound_slots : `None | int`, Optional (Keyword only)
            Additionally granted soundboard sound slots by the enhancement.
        
        additional_sticker_slots : `None | int`, Optional (Keyword only)
            Additionally granted sticker by the enhancement.
        
        features : ``None | str | GuildFeature| iterable<str> | iterable<GuildFeature>``, Optional (Keyword only)
            The granted features by the enhancement.
            
        """
        # additional_emoji_slots
        if additional_emoji_slots is ...:
            additional_emoji_slots = 0
        else:
            additional_emoji_slots = validate_additional_emoji_slots(additional_emoji_slots)
        
        # additional_soundboard_sound_slots
        if additional_soundboard_sound_slots is ...:
            additional_soundboard_sound_slots = 0
        else:
            additional_soundboard_sound_slots = validate_additional_soundboard_sound_slots(additional_soundboard_sound_slots)
        
        # additional_sticker_slots
        if additional_sticker_slots is ...:
            additional_sticker_slots = 0
        else:
            additional_sticker_slots = validate_additional_sticker_slots(additional_sticker_slots)
        
        # features
        if features is ...:
            features = None
        else:
            features = validate_features(features)
        
        # Construct
        self = object.__new__(cls)
        self.additional_emoji_slots = additional_emoji_slots
        self.additional_soundboard_sound_slots = additional_soundboard_sound_slots
        self.additional_sticker_slots = additional_sticker_slots
        self.features = features
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        field_added = False
        
        # additional_emoji_slots
        additional_emoji_slots = self.additional_emoji_slots
        if additional_emoji_slots:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' additional_emoji_slots = ')
            repr_parts.append(repr(additional_emoji_slots))
        
        # additional_soundboard_sound_slots
        additional_soundboard_sound_slots = self.additional_soundboard_sound_slots
        if additional_soundboard_sound_slots:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' additional_soundboard_sound_slots = ')
            repr_parts.append(repr(additional_soundboard_sound_slots))
        
        # additional_sticker_slots
        additional_sticker_slots = self.additional_sticker_slots
        if additional_sticker_slots:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' additional_sticker_slots = ')
            repr_parts.append(repr(additional_sticker_slots))
        
        # features
        features = self.features
        if (features is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' features = ')
            repr(features)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # additional_emoji_slots
        hash_value ^= self.additional_emoji_slots
        
        # additional_soundboard_sound_slots
        hash_value ^= self.additional_soundboard_sound_slots << 8
        
        # additional_sticker_slots
        hash_value ^= self.additional_sticker_slots << 16
        
        # features
        features = self.features
        if (features is not None):
            hash_value ^= hash(features)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # additional_emoji_slots
        if self.additional_emoji_slots != other.additional_emoji_slots:
            return False
        
        # additional_soundboard_sound_slots
        if self.additional_soundboard_sound_slots != other.additional_soundboard_sound_slots:
            return False
        
        # additional_sticker_slots
        if self.additional_sticker_slots != other.additional_sticker_slots:
            return False
        
        # features
        if self.features != other.features:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a SKU enhancement guild from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received SKU enhancement guild data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.additional_emoji_slots = parse_additional_emoji_slots(data)
        self.additional_soundboard_sound_slots = parse_additional_soundboard_sound_slots(data)
        self.additional_sticker_slots = parse_additional_sticker_slots(data)
        self.features = parse_features(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the SKU enhancement guild to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_additional_emoji_slots(self.additional_emoji_slots, data, defaults)
        put_additional_soundboard_sound_slots(self.additional_soundboard_sound_slots, data, defaults)
        put_additional_sticker_slots(self.additional_sticker_slots, data, defaults)
        put_features(self.features, data, defaults)
        
        return data
    
    
    def copy(self):
        """
        Copies the SKU enhancement guild.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.additional_emoji_slots = self.additional_emoji_slots
        new.additional_soundboard_sound_slots = self.additional_soundboard_sound_slots
        new.additional_sticker_slots = self.additional_sticker_slots
        new.features = self.features
        return new
    
    
    def copy_with(
        self, 
        *,
        additional_emoji_slots = ...,
        additional_soundboard_sound_slots = ...,
        additional_sticker_slots = ...,
        features = ...,
    ):
        """
        Copies the SKU enhancement guild with the given fields.
        
        Parameters
        ----------
        additional_emoji_slots : `None | int`, Optional (Keyword only)
            Additional granted emoji slots by the enhancement.
        
        additional_soundboard_sound_slots : `None | int`, Optional (Keyword only)
            Additionally granted soundboard sound slots by the enhancement.
        
        additional_sticker_slots : `None | int`, Optional (Keyword only)
            Additionally granted sticker by the enhancement.
        
        features : ``None | str | GuildFeature| iterable<str> | iterable<GuildFeature>``, Optional (Keyword only)
            The granted features by the enhancement.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # additional_emoji_slots
        if additional_emoji_slots is ...:
            additional_emoji_slots = self.additional_emoji_slots
        else:
            additional_emoji_slots = validate_additional_emoji_slots(additional_emoji_slots)
        
        # additional_soundboard_sound_slots
        if additional_soundboard_sound_slots is ...:
            additional_soundboard_sound_slots = self.additional_soundboard_sound_slots
        else:
            additional_soundboard_sound_slots = validate_additional_soundboard_sound_slots(additional_soundboard_sound_slots)
        
        # additional_sticker_slots
        if additional_sticker_slots is ...:
            additional_sticker_slots = self.additional_sticker_slots
        else:
            additional_sticker_slots = validate_additional_sticker_slots(additional_sticker_slots)
        
        # features
        if features is ...:
            features = self.features
        else:
            features = validate_features(features)
        
        # Construct
        new = object.__new__(type(self))
        new.additional_emoji_slots = additional_emoji_slots
        new.additional_soundboard_sound_slots = additional_soundboard_sound_slots
        new.additional_sticker_slots = additional_sticker_slots
        new.features = features
        return new
    
    
    def iter_features(self):
        """
        Iterates over the features granted by the enhancement.
        
        This method is an iterable generator.
        
        Yields
        ------
        feature : ``GuildFeature``
        """
        features = self.features
        if (features is not None):
            yield from features
    
    
    def has_feature(self, feature):
        """
        Returns whether the enhancement grants the given feature.
        
        Parameters
        ----------
        feature : ``GuildFeature``
            The feature to look for.
        
        Returns
        -------
        has_feature : `bool`
        """
        features = self.features
        if features is None:
            return False
        
        return feature in features
    
    



__all__ = ('GuildEnhancementEntitlementsDeleteEvent',)


from ..guild_enhancement_entitlements_create_event import GuildEnhancementEntitlementsCreateEvent

class GuildEnhancementEntitlementsDeleteEvent(GuildEnhancementEntitlementsCreateEvent):
    """
    Represents a `GUILD_POWERUP_ENTITLEMENTS_DELETE` event.
    
    Attributes
    ----------
    entitlements : ``None | tuple<Entitlement>``
        The affected entitlements.
    
    guild_id : `int`
        The guild's identifier where the event is for.
    """
    __slots__ = ()



__all__ = ()

from warnings import warn

from scarletio import Compound
from scarletio.web_common import FormData

from ...application import Application
from ...bases import maybe_snowflake
from ...builder.serialization import create_serializer
from ...builder.serialization_configuration import SerializationConfiguration
from ...component import ButtonStyle, InteractionForm
from ...http import DiscordApiClient
from ...interaction import InteractionEvent, InteractionResponseContext, InteractionResponseType, InteractionType
from ...message import Message, MessageFlag
from ...message.message.utils import try_resolve_interaction_message
from ...message.message_builder import (
    MessageBuilderInteractionComponentEdit, MessageBuilderInteractionFollowupCreate,
    MessageBuilderInteractionFollowupEdit, MessageBuilderInteractionResponseCreate,
    MessageBuilderInteractionResponseEdit
)

from ..functionality_helpers import application_command_autocomplete_choice_parser
from ..request_helpers import get_message_id

MESSAGE_FLAG_VALUE_INVOKING_USER_ONLY = MessageFlag().update_by_keys(invoking_user_only = True)


MESSAGE_SERIALIZER_INTERACTION_FOLLOWUP_CREATE = create_serializer(
    MessageBuilderInteractionFollowupCreate,
    SerializationConfiguration(
        [
            MessageBuilderInteractionFollowupCreate.allowed_mentions,
            MessageBuilderInteractionFollowupCreate.attachments,
            MessageBuilderInteractionFollowupCreate.components,
            MessageBuilderInteractionFollowupCreate.content,
            MessageBuilderInteractionFollowupCreate.embeds,
            MessageBuilderInteractionFollowupCreate.flags,
            MessageBuilderInteractionFollowupCreate.poll,
            MessageBuilderInteractionFollowupCreate.show_for_invoking_user_only,
            MessageBuilderInteractionFollowupCreate.tts,
            MessageBuilderInteractionFollowupCreate.voice_attachment,
        ],
        False,
    )
)

MESSAGE_SERIALIZER_INTERACTION_FOLLOWUP_EDIT = create_serializer(
    MessageBuilderInteractionFollowupEdit,
    SerializationConfiguration(
        [
            MessageBuilderInteractionFollowupEdit.allowed_mentions,
            MessageBuilderInteractionFollowupEdit.attachments,
            MessageBuilderInteractionFollowupEdit.components,
            MessageBuilderInteractionFollowupEdit.content,
            MessageBuilderInteractionFollowupEdit.embeds,
            MessageBuilderInteractionFollowupEdit.flags,
        ],
        True,
    )
)

MESSAGE_SERIALIZER_INTERACTION_COMPONENT_EDIT = create_serializer(
    MessageBuilderInteractionComponentEdit,
    SerializationConfiguration(
        [
            MessageBuilderInteractionComponentEdit.allowed_mentions,
            MessageBuilderInteractionComponentEdit.components,
            MessageBuilderInteractionComponentEdit.content,
            MessageBuilderInteractionComponentEdit.embeds,
            MessageBuilderInteractionComponentEdit.flags,
        ],
        True,
    )
)

MESSAGE_SERIALIZER_INTERACTION_RESPONSE_CREATE = create_serializer(
    MessageBuilderInteractionResponseCreate,
    SerializationConfiguration(
        [
            MessageBuilderInteractionResponseCreate.allowed_mentions,
            MessageBuilderInteractionResponseCreate.components,
            MessageBuilderInteractionResponseCreate.content,
            MessageBuilderInteractionResponseCreate.embeds,
            MessageBuilderInteractionResponseCreate.flags,
            MessageBuilderInteractionResponseCreate.poll,
            MessageBuilderInteractionResponseCreate.tts,
        ],
        True,
    )
)


MESSAGE_SERIALIZER_INTERACTION_RESPONSE_EDIT = create_serializer(
    MessageBuilderInteractionResponseEdit,
    SerializationConfiguration(
        [
            MessageBuilderInteractionResponseEdit.allowed_mentions,
            MessageBuilderInteractionResponseEdit.attachments,
            MessageBuilderInteractionResponseEdit.components,
            MessageBuilderInteractionResponseEdit.content,
            MessageBuilderInteractionResponseEdit.embeds,
            MessageBuilderInteractionResponseEdit.flags,
            MessageBuilderInteractionResponseEdit.poll,
            MessageBuilderInteractionResponseEdit.voice_attachment,
        ],
        True,
    )
)


def _assert__interaction_event_type(interaction_event):
    """
    Asserts whether the interaction event's type is correct.
    
    Parameters
    ----------
    interaction_event : ``InteractionEvent``
        Respective interaction event.
    
    Raises
    ------
    AssertionError
        - If `interaction_event` is not ``InteractionEvent``.
    """
    if not isinstance(interaction_event, InteractionEvent):
        raise AssertionError(
            f'`interaction` can be `{InteractionEvent.__name__}`, got '
            f'{interaction_event.__class__.__name__}; {interaction_event!r}.'
        )
    
    return True


def _assert__application_id(application_id):
    """
    Asserts whether the client's application is synced by checking the `application_id`'s value.
    
    Parameters
    ----------
    application_id : `int`
        Client's application's id.
    
    Raises
    ------
    AssertionError
        - If the client's application is not yet synced..
    """
    if application_id == 0:
        raise AssertionError(
            'The client\'s application is not yet synced.'
        )
    
    return True


def _assert__form(form):
    """
    Asserts whether the form's type is correct.
    
    Parameters
    ----------
    form : ``InteractionForm``
        Respective form.
    
    Raises
    ------
    AssertionError
        - If `form` is not ``InteractionForm``.
    """
    if not isinstance(form, InteractionForm):
        raise AssertionError(
            f'`form` can be `{InteractionForm.__name__}`, got '
            f'{type(form).__name__}; {form!r}.'
        )
    
    return True


def _process_interaction_response(interaction_event, interaction_response_data):
    """
    Processes interaction event trying to create a message from it.
    
    Parameters
    ----------
    interaction_event : ``InteractionEvent``
        Interaction to acknowledge.
    
    interaction_response_data : `None | dict<str, object>`
        Interaction response data.
    
    Returns
    -------
    message : ``None | Message``
        Returns `None` on failure.
    """
    if (interaction_response_data is None):
        return None
    
    nested_data = interaction_response_data.get('resource', None)
    if (nested_data is None):
        return None
    
    message_data = nested_data.get('message', None)
    if (message_data is None):
        return None
    
    message = interaction_event.channel._create_new_message(message_data)
    try_resolve_interaction_message(message, interaction_event)
    return message


class ClientCompoundInteractionEndpoints(Compound):
    
    application : Application
    api : DiscordApiClient
    
    
    async def interaction_application_command_acknowledge(
        self, interaction_event, wait = True, *, show_for_invoking_user_only = False
    ):
        """
        Acknowledges the given application command interaction.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction to acknowledge
        wait : `bool` = `True`, Optional
            Whether the interaction should be ensured asynchronously.
        show_for_invoking_user_only : `bool` = `False`, Optional (Keyword only)
            Whether the sent message should only be shown to the invoking user. Defaults to `False`.
        
        Returns
        -------
        response_message : ``None | Message``
            Returns a message oon success.
            If `wait` was given as `False` then `None`.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the interaction is already timed or out or was used, you will get:
        
        ```
        DiscordException Not Found (404), code = 10062: Unknown interaction
        ```
        """
        assert _assert__interaction_event_type(interaction_event)
        
        # Do not ack twice
        if not interaction_event.is_unanswered():
            return
        
        data = {'type': InteractionResponseType.source.value}
        
        if show_for_invoking_user_only:
            data['data'] = {'flags': MESSAGE_FLAG_VALUE_INVOKING_USER_ONLY}
        
        context = InteractionResponseContext(interaction_event, True, show_for_invoking_user_only)
        coroutine = self.api.interaction_response_message_create(
            interaction_event.id, interaction_event.token, data, {'with_response': True}
        )
        
        if not wait:        
            await context.ensure(coroutine, _process_interaction_response)
            return
        
        async with context:
            interaction_response_data = await coroutine
        
        # Example response
        # {
        #     'interaction': {
        #         'id': '...',
        #         'type': 2,
        #         'response_message_id': '...',
        #         'response_message_loading': True,
        #         'response_message_ephemeral': False,
        #         'channel_id': '...',
        #         'guild_id': '...'
        #     },
        #     'resource': {
        #         'type': 4,
        #         'message': {...},
        #      }
        #  }
        
        return _process_interaction_response(interaction_event, interaction_response_data)
    
    
    async def interaction_application_command_autocomplete(self, interaction_event, choices):
        """
        Forwards auto completion choices for the user.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction to acknowledge.
        
        choices : `None | iterable<str | int | float | (str, str | int | float)>
            Choices to show for the user.
        
        Raises
        ------
        TypeError
            If `choice` is neither `None` nor `iterable`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the interaction is already timed or out or was used, you will get:
        
        ```
        DiscordException Not Found (404), code = 10062: Unknown interaction
        ```
        """
        assert _assert__interaction_event_type(interaction_event)
        
        # Do not auto complete twice
        if not interaction_event.is_unanswered():
            return
        
        choices = application_command_autocomplete_choice_parser(choices)
        
        data = {
            'type': InteractionResponseType.application_command_autocomplete_result.value,
            'data': {
                'choices': choices,
            },
        }
        
        async with InteractionResponseContext(interaction_event, True, False):
            await self.api.interaction_response_message_create(
                interaction_event.id, interaction_event.token, data, None
            )
        
        # Example output:
        # {
        #     'interaction': {
        #         'id': '...',
        #         'type': 4,
        #         'channel_id': '...',
        #         'guild_id': '...',
        #     }
        # }
        # No need to handle
    
    
    async def interaction_form_send(self, interaction_event, form):
        """
        Responds on an interaction with a form.
        
        This function is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction to respond to.
        form : ``InteractionForm``
            The to respond with.
        
        Raises
        ------
        RuntimeError
            If cannot respond with a form on the given `interaction`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        Discord do not returns message data, so the method cannot return a ``Message`` either.
        
        If the interaction is already timed or out or was used, you will get:
        
        ```
        DiscordException Not Found (404), code = 10062: Unknown interaction
        ```
        """
        assert _assert__interaction_event_type(interaction_event)
        
        if not interaction_event.is_unanswered():
            warn(
                (
                    f'`{type(self).__name__}.interaction_response_form` called on an interaction already '
                    f'acknowledged / answered: {interaction_event!r}. Returning `None`.'
                ),
                ResourceWarning,
                stacklevel = 2,
            )
            
            return None
        
        assert _assert__form(form)
        
        if (
            (interaction_event.type is not InteractionType.application_command) and
            (interaction_event.type is not InteractionType.message_component)
        ):
            raise RuntimeError(
                f'Only `application_command` and `message_component` interactions can be answered with '
                f'form, got `{interaction_event.type.name}`; {interaction_event!r}; form = {form!r}.'
            )
        
        # Build payload
        data = {
            'data': form.to_data(),
            'type': InteractionResponseType.form.value,
        }
        
        async with InteractionResponseContext(interaction_event, False, True):
            await self.api.interaction_response_message_create(
                interaction_event.id, interaction_event.token, data, None
            )
            
        # Example response:
        # {
        #     'interaction': {
        #         'id': '...',
        #         'type': 2,
        #         'channel_id': '...',
        #         'guild_id': '...',
        #     }
        # }
        # No need to handle.
        
        return None
    
    
    async def interaction_response_message_create(
        self, interaction_event, *positional_parameters, **keyword_parameters,
    ):
        """
        Sends an interaction response. After receiving an ``InteractionEvent``, you should acknowledge it within
        `3` seconds to perform followup actions.
        
        Not like ``.message_create``, this endpoint can be called without any content to still acknowledge the
        interaction event. This method also wont return a ``Message`` object (thank to Discord), but at least
        ``.interaction_followup_message_create`` will. To edit or delete this message, you can use
        ``.interaction_response_message_edit`` and ``.interaction_response_message_delete``.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction to respond to.
        
        *positional_parameters : Positional parameters
            Additional parameters to create the message with.
        
        **keyword_parameters : Keyword parameters
            Additional parameters to create the message with.

        Other Parameters
        ----------------
        allowed_mentions : `None`,  ``AllowedMentionProxy``, `str`, ``UserBase``, ``Role``, `list` of \
                (`str`, ``UserBase``, ``Role`` ) , Optional
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions``
            for details.
        
        components : `None`, ``Component``, `(tuple | list)<Component, (tuple | list)<Component>>`
            Components attached to the message.
        
        content : `None`, `str`, Optional
            The message's content if given.
        
        embed : `None`, `Embed`, Optional
            Alternative for `embeds`.
        
        embeds : `None`, `list<Embed>`, Optional
            The new embedded content of the message.
        
        flags : `int`, ``MessageFlag`, Optional
            The message's flags.
        
        poll : `None`, ``Poll``, Optional
            The message's poll.
        
        show_for_invoking_user_only : `bool` = `False`, Optional (Keyword only)
            Whether the sent message should only be shown to the invoking user.
        
        silent : `bool` = `False`, Optional (Keyword only)
            Whether the message should be delivered silently.
        
        suppress_embeds : `bool` = `False`, Optional (Keyword only)
            Whether the message's embeds should be suppressed initially.
        
        tts : `bool` = `False`, Optional (Keyword only)
            Whether the message is text-to-speech.
        
        Returns
        -------
        response_message : ``None | Message``
            Returns a message on success.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the interaction is already timed or out or was used, you will get:
        
        ```
        DiscordException Not Found (404), code = 10062: Unknown interaction
        ```
        """
        assert _assert__interaction_event_type(interaction_event)
        
        message_data = MESSAGE_SERIALIZER_INTERACTION_RESPONSE_CREATE(positional_parameters, keyword_parameters)
        
        data = {}
        if message_data:
            data['data'] = message_data
        
        flags = message_data.get('flags', -1)
        
        # is_deferring
        if len(message_data) - (flags != -1):
            is_deferring = False
        else:
            is_deferring = True
        
        # show_for_invoking_user_only
        if flags != -1 and flags & MESSAGE_FLAG_VALUE_INVOKING_USER_ONLY:
            show_for_invoking_user_only = True
        else:
            show_for_invoking_user_only = False
        
        if is_deferring:
            response_type = InteractionResponseType.source
        else:
            response_type = InteractionResponseType.message_and_source
        
        data['type'] = response_type.value
        
        async with InteractionResponseContext(interaction_event, is_deferring, show_for_invoking_user_only):
            interaction_response_data = await self.api.interaction_response_message_create(
                interaction_event.id, interaction_event.token, data, {'with_response': True}
            )
            
        # Example output:
        # {
        #     'interaction': {
        #         'id': '1287034694394712105',
        #         'type': 2,
        #         'response_message_id': '...',
        #         'response_message_loading': False,
        #         'response_message_ephemeral': False,
        #         'channel_id': '...',
        #         'guild_id': '...'
        #     },
        #     'resource': {
        #         'type': 4,
        #         'message': {...},
        #     },
        # }
        
        output = _process_interaction_response(interaction_event, interaction_response_data)
        return output
    
    
    async def interaction_component_acknowledge(self, interaction_event, wait = True):
        """
        Acknowledges the given component interaction.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction to acknowledge
        wait : `bool` = `True`, Optional
            Whether the interaction should be ensured asynchronously.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the interaction is already timed or out or was used, you will get:
        
        ```
        DiscordException Not Found (404), code = 10062: Unknown interaction
        ```
        """
        assert _assert__interaction_event_type(interaction_event)
        
        # Do not ack twice
        if not interaction_event.is_unanswered():
            return
        
        data = {'type': InteractionResponseType.component.value}
        
        context = InteractionResponseContext(interaction_event, True, False)
        coroutine = self.api.interaction_response_message_create(
            interaction_event.id, interaction_event.token, data, None
        )
        
        
        if wait:
            async with context:
                await coroutine
        else:
            await context.ensure(coroutine)
        
        # Example output:
        # {
        #     'interaction': {
        #         'id': '...',
        #         'type': 3,
        #         'response_message_id': '...',
        #         'response_message_loading': False,
        #         'response_message_ephemeral': False,
        #         'channel_id': '...',
        #         'guild_id': '...',
        #     }
        # }
        # No need to handle.
    
    
    async def interaction_embedded_activity_launch(self, interaction_event, wait = True):
        """
        Acknowledges the interaction by launching an embedded activity. Omits the option for response message, but
        allowed followup messages as normal.
        
        Can be used if: `interaction_event.target_type is ApplicationCommandTargetType.embedded_activity_launch`.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction to launch its embedded activity of.
        
        wait : `bool` = `True`, Optional
            Whether the interaction should be ensured asynchronously.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the interaction is already timed or out or was used, you will get:
        
        ```
        DiscordException Not Found (404), code = 10062: Unknown interaction
        ```
        """
        assert _assert__interaction_event_type(interaction_event)
        
        data = {
            'type': InteractionResponseType.embedded_activity_launch.value,
        }
        
        context = InteractionResponseContext(interaction_event, False, False)
        coroutine = self.api.interaction_response_message_create(
            interaction_event.id, interaction_event.token, data, None
        )
        
        if wait:
            async with context:
                await coroutine
        else:
            await context.ensure(coroutine)
        
        # Example output:
        # {
        #     'interaction': {
        #         'id': '...',
        #         'type': 2,
        #         'activity_instance_id': 'i-...-gc-...-...',
        #         'channel_id': '...',
        #         'guild_id': '...',
        #     },
        #     'resource': {
        #         'type': 12,
        #         'activity_instance': {
        #             'id': 'i-...-gc-...-...',
        #          },
        #      },
        #  }
        # No need to handle.
    
    
    async def interaction_response_message_edit(
        self, interaction_event, *positional_parameters, **keyword_parameters,
    ):
        """
        Edits the given `interaction`'s source response. If the source interaction event was only deferred, this call
        will send the message as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction, what's source response message will be edited.
        
        *positional_parameters : Positional parameters
            Additional parameters to edit the message with.
        
        **keyword_parameters : Keyword parameters
            Additional parameters to edit the message with.
        
        Other Parameters
        ----------------
        allowed_mentions : `None`,  ``AllowedMentionProxy``, `str`, ``UserBase``, ``Role``, `list` of \
                (`str`, ``UserBase``, ``Role`` ) , Optional
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions``
            for details.
        
        attachments : `None | object`, Optional (Keyword only)
            Attachments to send.
        
        components : `None`, ``Component``, `(tuple | list)<Component, (tuple | list)<Component>>`
            Components attached to the message.
            
            Pass it as `None` remove the actual ones.
        
        content : `None`, `str`, Optional
            The new content of the message.
        
        embed : `None`, `Embed`, Optional
            Alternative for `embeds`.
        
        embeds : `None`, `list<Embed>`, Optional
            The new embedded content of the message.
            
            By passing it as `None`, you can remove the old.
        
        file : `None | object`, Optional (Keyword only)
            Alternative for `attachments`.
        
        files : `None | object`, Optional (Keyword only)
            Alternative for `attachments`.
        
        flags : `int`, ``MessageFlag`, Optional
            The message's new flags.
        
        poll : `None`, ``Poll``, Optional
            The message's poll.
        
        suppress_embeds : `bool`, Optional (Keyword only)
            Whether the message's embeds should be suppressed or unsuppressed.
        
        voice_attachment : ``None | VoiceAttachment``, Optional (Keyword only)
            Modifies the message to be a voice message, allowing it to contain just a single voice attachment.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        message_data = MESSAGE_SERIALIZER_INTERACTION_RESPONSE_EDIT(positional_parameters, keyword_parameters)
        if not message_data:
            return
        
        async with InteractionResponseContext(interaction_event, False, False):
            await self.api.interaction_response_message_edit(
                application_id, interaction_event.id, interaction_event.token, message_data
            )
    
    
    async def interaction_component_message_edit(
        self, interaction_event, *positional_parameters, **keyword_parameters
    ):
        """
        Edits the given component interaction's source message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction, what's source response message will be edited.
        
        *positional_parameters : Positional parameters
            Additional parameters to edit the message with.
        
        **keyword_parameters : Keyword parameters
            Additional parameters to edit the message with.
        
        Other Parameters
        ----------------
        allowed_mentions : `None`,  ``AllowedMentionProxy``, `str`, ``UserBase``, ``Role``, `list` of \
                (`str`, ``UserBase``, ``Role`` ) , Optional
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions``
            for details.
        
        components : `None`, ``Component``, `(tuple | list)<Component, (tuple | list)<Component>>`
            Components attached to the message.
            
            Pass it as `None` remove the actual ones.
        
        content : `None`, `str`, Optional
            The new content of the message.
        
        embed : `None`, `Embed`, Optional
            Alternative for `embeds`.
        
        embeds : `None`, `list<Embed>`, Optional
            The new embedded content of the message.
            
            By passing it as `None`, you can remove the old.
        
        flags : `int`, ``MessageFlag`, Optional
            The message's new flags.
        
        suppress_embeds : `bool`, Optional (Keyword only)
            Whether the message's embeds should be suppressed or unsuppressed.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__interaction_event_type(interaction_event)
        
        message_data = MESSAGE_SERIALIZER_INTERACTION_COMPONENT_EDIT(positional_parameters, keyword_parameters)
        
        if message_data:
            data = {
                'data': message_data,
                'type': InteractionResponseType.component_message_edit.value,
            }
            deferring = False
        
        else:
            data = {
                'type': InteractionResponseType.component.value,
            }
            deferring = True
        
        
        async with InteractionResponseContext(interaction_event, deferring, False):
            await self.api.interaction_response_message_create(
                interaction_event.id, interaction_event.token, data, None
            )
        
        # Example output:
        # {
        #     'interaction': {
        #         'id': '...',
        #         'type': 3,
        #         'response_message_id': '...',
        #         'response_message_loading': False,
        #         'response_message_ephemeral': False,
        #         'channel_id': '...',
        #         'guild_id': '...'
        #     },
        #     'resource': {
        #         'type': 7,
        #         'message': {...},
        #     }
        # }
    
    
    async def interaction_response_message_delete(self, interaction_event):
        """
        Deletes the given `interaction`'s source response message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction, what's source response message will be deleted.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        await self.api.interaction_response_message_delete(application_id, interaction_event.id, interaction_event.token)
    
    
    async def interaction_response_message_get(self, interaction_event):
        """
        Gets the given `interaction`'s source response message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction, what's source response message will be deleted.
        
        Returns
        -------
        message : ``Message``
            The created message.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        
        message_data = await self.api.interaction_response_message_get(
            application_id, interaction_event.id, interaction_event.token
        )
        
        return Message.from_data(message_data)
    
    
    async def interaction_followup_message_create(
        self, interaction_event, *positional_parameters, **keyword_parameters
    ):
        """
        Sends a followup message with the given interaction.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction to create followup message with.
        
        *positional_parameters : Positional parameters
            Additional parameters to create the message with.
        
        **keyword_parameters : Keyword parameters
            Additional parameters to create the message with.
        
        Other Parameters
        ----------------
        allowed_mentions : `None`,  ``AllowedMentionProxy``, `str`, ``UserBase``, ``Role``, `list` of \
                (`str`, ``UserBase``, ``Role`` ) , Optional
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions``
            for details.
        
        attachments : `None | object`, Optional (Keyword only)
            Attachments to send.
        
        components : `None`, ``Component``, `(tuple | list)<Component, (tuple | list)<Component>>`
            Components attached to the message.
        
        content : `None`, `str`, Optional
            The message's content if given.
        
        poll : `None`, ``Poll``, Optional
            The message's poll.
            
            > Response message must be created or else discord will ignore the `poll` field.
        
        embed : `None`, `Embed`, Optional
            Alternative for `embeds`.
        
        embeds : `None`, `list<Embed>`, Optional
            The new embedded content of the message.
        
        file : `None | object`, Optional (Keyword only)
            Alternative for `attachments`.
        
        files : `None | object`, Optional (Keyword only)
            Alternative for `attachments`.
        
        flags : `int`, ``MessageFlag`, Optional
            The message's flags.
        
        show_for_invoking_user_only : `bool` = `False`, Optional (Keyword only)
            Whether the sent message should only be shown to the invoking user.
        
        silent : `bool` = `False`, Optional (Keyword only)
            Whether the message should be delivered silently.
        
        suppress_embeds : `bool` = `False`, Optional (Keyword only)
            Whether the message's embeds should be suppressed initially.
        
        tts : `bool` = `False`, Optional (Keyword only)
            Whether the message is text-to-speech.
        
        voice_attachment : ``None | VoiceAttachment``, Optional (Keyword only)
            Modifies the message to be a voice message, allowing it to contain just a single voice attachment.
        
        Returns
        -------
        message : ``None | Message``
            Returns `None` if there is nothing to send.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        message_data = MESSAGE_SERIALIZER_INTERACTION_FOLLOWUP_CREATE(positional_parameters, keyword_parameters)
        if not message_data:
            return
        
        # show_for_invoking_user_only
        if isinstance(message_data, FormData):
            flags = message_data.fields[0].value.get('flags', 0)
        else:
            flags = message_data.get('flags', 0)
        if flags & MESSAGE_FLAG_VALUE_INVOKING_USER_ONLY:
            show_for_invoking_user_only = True
        else:
            show_for_invoking_user_only = False
        
        
        async with InteractionResponseContext(interaction_event, False, show_for_invoking_user_only):
            message_data = await self.api.interaction_followup_message_create(
                application_id, interaction_event.id, interaction_event.token, message_data
            )
        
        message = interaction_event.channel._create_new_message(message_data)
        try_resolve_interaction_message(message, interaction_event)
        return message
    
    
    async def interaction_followup_message_edit(
        self, interaction_event, message, *positional_parameters, **keyword_parameters,
    ):
        """
        Edits the given interaction followup message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction with what the followup message was sent with.
        
        message : ``Message``, `int`
            The interaction followup's message to edit.
        
        *positional_parameters : Positional parameters
            Additional parameters to edit the message with.
        
        **keyword_parameters : Keyword parameters
            Additional parameters to edit the message with.
        
        Other Parameters
        ----------------
        allowed_mentions : `None`,  ``AllowedMentionProxy``, `str`, ``UserBase``, ``Role``, `list` of \
                (`str`, ``UserBase``, ``Role`` ) , Optional
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions``
            for details.
        
        attachments : `None | object`, Optional (Keyword only)
            Attachments to send.
        
        components : `None`, ``Component``, `(tuple | list)<Component, (tuple | list)<Component>>`
            Components attached to the message.
            
            Pass it as `None` remove the actual ones.
        
        content : `None`, `str`, Optional
            The new content of the message.
        
        embed : `None`, `Embed`, Optional
            Alternative for `embeds`.
        
        embeds : `None`, `list<Embed>`, Optional
            The new embedded content of the message.
            
            By passing it as `None`, you can remove the old.
        
        file : `None | object`, Optional (Keyword only)
            Alternative for `attachments`.
        
        files : `None | object`, Optional (Keyword only)
            Alternative for `attachments`.
        
        flags : `int`, ``MessageFlag`, Optional
            The message's new flags.
        
        suppress_embeds : `bool`, Optional (Keyword only)
            Whether the message's embeds should be suppressed or unsuppressed.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        # Detect message id
        # 1.: Message
        # 2.: int (str)
        # 5.: raise
        
        if isinstance(message, Message):
            message_id = message.id
        else:
            message_id = maybe_snowflake(message)
            if (message_id is None):
                raise TypeError(
                    f'`message` can be `{Message.__name__}`, `int`, got '
                    f'{type(message).__name__}, {message!r}.'
                )
        
        message_data = MESSAGE_SERIALIZER_INTERACTION_FOLLOWUP_EDIT(positional_parameters, keyword_parameters)
        if not message_data:
            return
        
        async with InteractionResponseContext(interaction_event, False, False):
            # We receive the new message data, but we do not update the message, so dispatch events can get the
            # difference.
            await self.api.interaction_followup_message_edit(
                application_id, interaction_event.id, interaction_event.token, message_id, message_data
            )
    
    
    async def interaction_followup_message_delete(self, interaction_event, message):
        """
        Deletes an interaction's followup message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction with what the followup message was sent with.
        message : ``Message``, `int`
            The interaction followup's message to edit.
        
        Raises
        ------
        TypeError
            If `message` was not given neither as ``Message``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        # Detect message id
        # 1.: Message
        # 2.: int (str)
        # 5.: raise
        
        if isinstance(message, Message):
            message_id = message.id
        else:
            message_id = maybe_snowflake(message)
            if (message_id is None):
                raise TypeError(
                    f'`message` can be `{Message.__name__}`, `int`, got {message.__class__.__name__}; {message!r}.'
                )
        
        await self.api.interaction_followup_message_delete(
            application_id, interaction_event.id, interaction_event.token, message_id
        )
    
    
    async def interaction_followup_message_get(self, interaction_event, message):
        """
        Gets a previously sent message with an interaction.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction with what the followup message was sent with.
        message : ``Message``, `int`
            The message or it's identifier to get.
        
        Returns
        -------
        message : ``Message``
        
        Raises
        ------
        TypeError
            - If `message_id` was not given as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        message_id = get_message_id(message)
        
        message_data = await self.api.interaction_followup_message_get(
            application_id, interaction_event.id, interaction_event.token, message_id
        )
        
        return Message.from_data(message_data)
    
    
    async def interaction_require_subscription(self, interaction_event):
        """
        Requires the user to subscribe to the application.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Interaction to respond to.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if not interaction_event.is_unanswered():
            warn(
                (
                    f'`{type(self).__name__}.interaction_require_subscription` is deprecated. '
                    f'Please use a component with `{ButtonStyle.__name__}.subscription` instead.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
        
        _assert__interaction_event_type(interaction_event)
        
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        data = {'type': InteractionResponseType.require_subscription.value}
        
        async with InteractionResponseContext(interaction_event, False, True):
            # Uses the same endpoint as message create
            await self.api.interaction_response_message_create(
                interaction_event.id, interaction_event.token, data, None
            )



__all__ = ('InteractionComponentMetadataSection',)

from scarletio import copy_docs

from .base import InteractionComponentMetadataBase
from .fields import (
    parse_components, parse_thumbnail, put_components, put_thumbnail, validate_components__section, validate_thumbnail
)


class InteractionComponentMetadataSection(InteractionComponentMetadataBase):
    """
    Interaction component metadata representing a section component.
    
    Attributes
    ----------
    components : ``None | tuple<InteractionComponent>``
        Sub-components nested inside.
    
    thumbnail : ``None | InteractionComponent``
        The thumbnail or other accessory (button) of a section component.
    """
    __slots__ = ('components', 'thumbnail')
    
    def __new__(cls, *, components = ..., thumbnail = ...):
        """
        Creates a new interaction component metadata from the given fields.
        
        Parameters
        ----------
        components : ``None | iterable<InteractionComponent>``, Optional (Keyword only)
            Sub-components nested inside.
        
        thumbnail : ``None | InteractionComponent``, Optional (Keyword only)
            The thumbnail or other accessory (button) of a section component.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # components
        if components is ...:
            components = None
        else:
            components = validate_components__section(components)
        
        # thumbnail
        if thumbnail is ...:
            thumbnail = None
        else:
            thumbnail = validate_thumbnail(thumbnail)
        
        # Construct.
        self = object.__new__(cls)
        self.components = components
        self.thumbnail = thumbnail
        return self
    
    
    @classmethod
    def from_keyword_parameters(cls, keyword_parameters):
        """
        Creates a new interaction component metadata from the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict<str, object>`
            Keyword parameters to build the metadata from.
        
        Returns
        -------
        self : `instance<type<cls>>`
        
        Raises
        ------
        TypeError
            - If a keyword parameter's type is incorrect.
        ValueError
            - If a keyword parameter's value is incorrect.
        """
        return cls(
            components = keyword_parameters.pop('components', ...),
            thumbnail = keyword_parameters.pop('thumbnail', ...),
        )
    
    
    @copy_docs(InteractionComponentMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        # components
        repr_parts.append(' components = ')
        repr_parts.append(repr(self.components))
        
        # thumbnail
        repr_parts.append(', thumbnail = ')
        repr_parts.append(repr(self.thumbnail))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(InteractionComponentMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # components
        components = self.components
        if (components is not None):
            hash_value ^= len(components)
            
            for component in components:
                hash_value ^= hash(component)
        
        # thumbnail
        thumbnail = self.thumbnail
        if (thumbnail is not None):
            hash_value ^= hash(thumbnail)
        
        return hash_value
    
    
    @copy_docs(InteractionComponentMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # components
        if self.components != other.components:
            return False
        
        # thumbnail
        if self.thumbnail != other.thumbnail:
            return False
        
        return True
    
    
    @copy_docs(InteractionComponentMetadataBase._match_to_component)
    def _match_to_component(self, other):
        # components
        self_components = self.components
        other_components = other.components
        
        if (self_components is None) != (other_components is None):
            return False
        
        if (
            (self_components is not None) and
            (
                (len(self_components) != len(other_components)) or
                (not all(
                    self_component % other_component
                    for self_component, other_component
                    in zip(self_components, other_components)
                ))
            )
        ):
            return False
        
        # thumbnail
        self_thumbnail = self.thumbnail
        other_thumbnail = other.thumbnail
        
        if (self_thumbnail is None) != (other_thumbnail is None):
            return False
        
        if (self_thumbnail is not None) and (not (self_thumbnail % other_thumbnail)):
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(InteractionComponentMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.components = parse_components(data)
        self.thumbnail = parse_thumbnail(data)
        return self
    
    
    @copy_docs(InteractionComponentMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_components(self.components, data, defaults)
        put_thumbnail(self.thumbnail, data, defaults)
        return data
    
    
    @copy_docs(InteractionComponentMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        # components
        components = self.components
        if (components is not None):
            components = (*(component.copy() for component in components),)
        new.components = components
        
        # thumbnail
        thumbnail = self.thumbnail
        if (thumbnail is not None):
            thumbnail = thumbnail.copy()
        new.thumbnail = thumbnail
        
        return new
    
    
    def copy_with(self, *, components = ..., thumbnail = ...):
        """
        Copies the interaction component metadata with the given fields.
        
        Parameters
        ----------
        components : ``None | iterable<InteractionComponent>``, Optional (Keyword only)
            Sub-components nested inside.
        
        thumbnail : ``None | InteractionComponent``, Optional (Keyword only)
            The thumbnail or other accessory (button) of a section component.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        # components
        if components is ...:
            components = self.components
            if (components is not None):
                components = (*(component.copy() for component in components),)
        else:
            components = validate_components__section(components)
        
        # thumbnail
        if thumbnail is ...:
            thumbnail = self.thumbnail
            if (thumbnail is not None):
                thumbnail = thumbnail.copy()
        else:
            thumbnail = validate_thumbnail(thumbnail)
        
        # Construct.
        new = object.__new__(type(self))
        new.components = components
        new.thumbnail = thumbnail
        return new
    
    
    @copy_docs(InteractionComponentMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            components = keyword_parameters.pop('components', ...),
            thumbnail = keyword_parameters.pop('thumbnail', ...),
        )
    
    
    @copy_docs(InteractionComponentMetadataBase.iter_custom_ids_and_values)
    def iter_custom_ids_and_values(self):
        components = self.components
        if (components is not None):
            for component in components:
                yield from component.iter_custom_ids_and_values()
        
        thumbnail = self.thumbnail
        if (thumbnail is not None):
            yield from thumbnail.iter_custom_ids_and_values()



__all__ = ('Event',)

from scarletio import CallableAnalyzer, RichAttributeErrorBaseType

from .event_deprecation import EventDeprecation


def _validate_parameter_count(parameter_count):
    """
    Validates the `parameter_count` parameter of ``Event.__new__``.
    
    Parameters
    ----------
    parameter_count : `object`
        The parameter to validate.
    
    Returns
    -------
    parameter_count : `int`
    
    Raises
    ------
    TypeError
    ValueError
    """
    # parameter_count
    if type(parameter_count) is int:
        pass
    elif isinstance(parameter_count, int):
        parameter_count = int(parameter_count)
    else:
        raise TypeError(
            f'`parameter_count` can be `int`, got {type(parameter_count).__name__}; {parameter_count!r}.'
        )
    
    if (parameter_count < 0):
        raise ValueError(
            f'`parameter_count` must be positive, got {parameter_count!r}.'
        )
    
    return parameter_count


def _validate_default_handler(default_handler, parameter_count):
    """
    Validates the given `default_handler`.
    
    Parameters
    ----------
    default_handler : `object`
        Default handler to add by default.
    
    parameter_count : `int`
        How much parameters does the event handler should accept.
    
    Returns
    -------
    default_handler : `None | async-callable`
        Default handler to add by default.
    
    instance_default_handler : `bool`
        Whether `default_handler` should be instanced.
    """
    while True:
        if (default_handler is None):
            instance_default_handler = False
            break
        
        analyzer = CallableAnalyzer(default_handler)
        
        if analyzer.is_async():
            min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
            if min_ <= parameter_count:
                if min_ == parameter_count:
                    instance_default_handler = False
                    break
                
                # min < expected
                if max_ >= parameter_count:
                    instance_default_handler = False
                    break
                
                if analyzer.accepts_args():
                    instance_default_handler = False
                    break
            
            raise TypeError(
                f'`default_handler` should accept `{parameter_count!r}` parameters as defined, '
                f'but it accepts [{min_!r}:{max_!r}], got {default_handler!r}.'
            )
        
        
        if analyzer.can_instance_to_async_callable():
            sub_analyzer = CallableAnalyzer(default_handler.__call__, as_method = True)
            if sub_analyzer.is_async():
                min_, max_ = sub_analyzer.get_non_reserved_positional_parameter_range()
                
                if min_ <= parameter_count:
                    if min_ == parameter_count:
                        instance_default_handler = True
                        break
                    
                    # min < expected
                    if max_ >= parameter_count:
                        instance_default_handler = True
                        break
                    
                    if sub_analyzer.accepts_args():
                        instance_default_handler = True
                        break
                    
                raise TypeError(
                    f'`default_handler` should accept `{parameter_count!r}` parameters as '
                    f'defined, but after instancing it accepts [{min_!r}:{max_!r}], got '
                    f'{default_handler!r}.'
                )
        
        raise TypeError(
            f'`default_handler` can be `None`, `async-callable`, instantiable to '
            f'`async-callable`, got {default_handler!r}.'
        )

    return default_handler, instance_default_handler


def _validate_deprecation(deprecation):
    """
    Validates the given `deprecation`
    
    Parameters
    ----------
    deprecation : `object`
        Deprecations to validate.
    
    Returns
    -------
    deprecation `None | tuple<Deprecations>`
    """
    if deprecation is None:
        return None
    
    if isinstance(deprecation, EventDeprecation):
        if deprecation.allowed:        
            return deprecation
        
        return None
    
    raise TypeError(
        f'`deprecation` can be `None`, `{EventDeprecation.__name__}`, '
        f'got {type(deprecation).__name__}; {deprecation!r}.'
    )


class Event(RichAttributeErrorBaseType):
    """
    Custom event slot for event handler plugins.
    
    Attributes
    ----------
    default_handler : `None | async-callable`
        Default handler to add by default.
    
    deprecation : `None | EventDeprecation`
        Deprecation for this event.
    
    instance_default_handler : `bool`
        Whether `default_handler` should be instanced.
    
    parameter_count : `int`
        How much parameters does the event handler should accept.
    """
    __slots__ = ('default_handler', 'deprecation', 'instance_default_handler', 'parameter_count')
    
    def __new__(cls, parameter_count, default_handler = None, *, deprecation = None):
        """
        Creates a new event handler instance form the given parameters.
        
        Parameters
        ---------
        parameter_count : `int`
            How much parameters does the event handler should accept.
        
        default_handler : `None | async-callable` = `None`, Optional
            Default handler to add by default.
    
        deprecation : `None | EventDeprecation` = `None`, Optional (Keyword only)
            Deprecation for this event.
        
        Raises
        ------
        TypeError
            - If `parameter_count` is not `int`.
            - If `default_handler` is neither `None`, `async-callable` and cannot be instanced to async callable either.
            - If `default_handler` accepts different amount of parameters as `parameter_count` defined.
        ValueError
            - If `parameter_count` is negative.
        """
        parameter_count = _validate_parameter_count(parameter_count)
        default_handler, instance_default_handler = _validate_default_handler(default_handler, parameter_count)
        deprecation = _validate_deprecation(deprecation)
        
        # Construct
        self = object.__new__(cls)
        self.default_handler = default_handler
        self.deprecation = deprecation
        self.instance_default_handler = instance_default_handler
        self.parameter_count = parameter_count
        return self
    
    
    def __repr__(self):
        """Returns the custom event's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # parameter_count
        repr_parts.append(' parameter_count = ')
        repr_parts.append(repr(self.parameter_count))
        
        # default_handler
        default_handler = self.default_handler
        if (default_handler is not None):
            repr_parts.append(', default_handler = ')
            repr_parts.append(repr(default_handler))
            
            if self.instance_default_handler:
                repr_parts.append(' (instance)')
        
        # deprecation
        deprecation = self.deprecation
        if (deprecation is not None):
            repr_parts.append(', deprecation = ')
            repr_parts.append(repr(deprecation))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two events are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # default_handler
        if self.default_handler != other.default_handler:
            return False
        
        # deprecation
        if self.deprecation != other.deprecation:
            return False
        
        # instance_default_handler
        if self.instance_default_handler != other.instance_default_handler:
            return False
        
        # parameter_count
        if self.parameter_count != other.parameter_count:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the event's hash value."""
        hash_value = 0
        
        # default_handler
        default_handler = self.default_handler
        if (default_handler is not None):
            try:
                default_handler_hash = hash(default_handler)
            except TypeError:
                default_handler_hash = object.__hash__(default_handler)
            hash_value ^= default_handler_hash
        
        # deprecation
        deprecation = self.deprecation
        if (deprecation is not None):
            hash_value ^= hash(deprecation)
        
        # instance_default_handler
        hash_value ^= self.instance_default_handler << 8
        
        # parameter_count
        hash_value ^= self.parameter_count
        
        return hash_value



__all__ = ()

from scarletio import Compound

from ...http import DiscordApiClient
from ...payload_building import build_create_payload, build_edit_payload
from ...scheduled_event import ScheduledEvent
from ...scheduled_event.scheduled_event.utils import (
    SCHEDULED_EVENT_CREATE_FIELD_CONVERTERS, SCHEDULED_EVENT_EDIT_FIELD_CONVERTERS,
    scheduled_event_occasion_overwrite_get
)
from ...scheduled_event.scheduled_event_occasion_overwrite.fields import validate_timestamp
from ...scheduled_event.scheduled_event_occasion_overwrite.utils import (
    SCHEDULED_EVENT_OCCASION_OVERWRITE_CREATE_FIELD_CONVERTERS, SCHEDULED_EVENT_OCCASION_OVERWRITE_EDIT_FIELD_CONVERTERS
)
from ...user import User
from ...utils import datetime_to_id, datetime_to_timestamp, log_time_converter

from ..request_helpers import (
    get_guild_id, get_scheduled_event_and_guild_id_and_id, get_scheduled_event_guild_id_and_id
)


class ClientCompoundScheduledEventEndpoints(Compound):
    
    api : DiscordApiClient
    
    
    async def scheduled_event_create(
        self, guild, scheduled_event_template = None, reason = None, **keyword_parameters
    ):
        """
        Creates a guild scheduled events.
        
        To define, where the event will take place, use the `location`, `stage`, `voice` parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild, where to create the scheduled event.
        
        scheduled_event_template : `None`, ``ScheduledEvent``` = `None`, Optional
            Scheduled event to use as a template.
        
        reason : `None | str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other Parameters
        ----------------
        description : `None | str` = `None`, Optional (Keyword only)
            The event's description. It's length can be in range [0:1000].
        
        end : `None | DateTime` = `None`, Optional (Keyword only)
            When the event will end.
        
        location : `None | str` = `None`, Optional (Keyword only)
            The location, where the event will take place.
        
        name : `str`, Optional (Keyword only)
            The event's name. It's length can be in range [1:100].
        
        privacy_level : ``PrivacyLevel``, `int` = `None`, Optional (Keyword only)
            The privacy level of the event. Whether it is global or guild only.
        
        schedule : ``None | Schedule`` = `None`, Optional (Keyword only)
            How the scheduled event should re-occur.
        
        stage : ``None | int | Channel`` = `None`, Optional (Keyword only)
            The stage channel, where the event will take place.
        
        start : `DateTime`, Optional (Keyword only)
            When the event will start.
        
        voice : ``None | int | Channel`` = `None`, Optional (Keyword only)
            The voice channel, where the event will take place.
        
        Returns
        -------
        scheduled_event : ``ScheduledEvent``
            The created scheduled event.
        
        Raises
        ------
        TypeError
            - If `guild` is neither ``Guild`` nor `int`.
            - Parameter of incorrect type given.
            - Extra parameters.
        ValueError
            - Parameter of incorrect value given.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        data = build_create_payload(
            scheduled_event_template, SCHEDULED_EVENT_CREATE_FIELD_CONVERTERS, keyword_parameters
        )
        scheduled_event_data = await self.api.scheduled_event_create(guild_id, data, reason)
        return ScheduledEvent.from_data(scheduled_event_data)
    
    
    async def scheduled_event_edit(
        self,
        scheduled_event,
        scheduled_event_template = None,
        *,
        reason = None,
        **keyword_parameters,
    ):
        """
        Edits the given scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        scheduled_event : ``ScheduledEvent | (int, int)``
            The scheduled event to edit.
        
        scheduled_event_template : `None`, ``ScheduledEvent``` = `None`, Optional
            Scheduled event to use as a template.
        
        reason : `None | str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other Parameters
        ----------------
        description : `None | str`, Optional (Keyword only)
            The new description of the scheduled event. It's length can be in range [0:1000].
        
        end : `None | DateTime`
            The end of the of the scheduled event.
        
        location : `str`, Optional (Keyword only)
            The new location, where the event will take place.
        
        name : `str`, Optional (Keyword only)
            The new name of the scheduled event. It's length can be in range [1:100].
        
        privacy_level : ``PrivacyLevel``, `int`, Optional (Keyword only)
            The privacy level of the event. Whether it is global or guild only.
        
        schedule : ``None | Schedule`` = `None`, Optional (Keyword only)
            How the scheduled event should re-occur.
        
        stage : ``Channel``, `int`, Optional (Keyword only)
            The new stage channel, where the event will take place.
        
        start : `DateTime`, Optional (Keyword only)
            The new start of the scheduled event.
        
        status : `str`, ``ScheduledEventStatus``, Optional (Keyword only)
            Thew new status of the scheduled event.
        
        voice : ``Channel``, `int`, Optional (Keyword only)
            The new voice channel, where the event will take place.
        
        Raises
        ------
        TypeError
            - If `scheduled_event` is invalid type.
            - Parameter of incorrect type given.
            - Extra parameters.
        ValueError
            - Parameter of incorrect value given.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        scheduled_event, guild_id, scheduled_event_id = get_scheduled_event_and_guild_id_and_id(scheduled_event)
        
        data = build_edit_payload(
            scheduled_event, scheduled_event_template, SCHEDULED_EVENT_EDIT_FIELD_CONVERTERS, keyword_parameters
        )
        if data:
            await self.api.scheduled_event_edit(guild_id, scheduled_event_id, data, reason)
    
    
    async def scheduled_event_delete(self, scheduled_event):
        """
        Edits the given scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        scheduled_event : ``ScheduledEvent | (int, int)``
            The scheduled event to edit.
        
        Raises
        ------
        TypeError
            If `scheduled_event` is invalid type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, scheduled_event_id = get_scheduled_event_guild_id_and_id(scheduled_event)
        await self.api.scheduled_event_delete(guild_id, scheduled_event_id)
    
    
    async def scheduled_event_get(self, scheduled_event, *, force_update = False):
        """
        Gets the given scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        scheduled_event : ``ScheduledEvent | (int, int)``
            The scheduled event to get.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the scheduled event should be requested even if it supposed to be up to date.
        
        Returns
        -------
        scheduled_event : ``ScheduledEvent``
        
        Raises
        ------
        TypeError
            If `scheduled_event` is invalid type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        scheduled_event, guild_id, scheduled_event_id = get_scheduled_event_and_guild_id_and_id(scheduled_event)
        if (scheduled_event is None) or force_update:
            data = await self.api.scheduled_event_get(guild_id, scheduled_event_id, {'with_user_count', None})
            
            scheduled_event, is_created = ScheduledEvent.from_data_is_created(data)
            if not is_created:
                scheduled_event._update_attributes(data)
        
        return scheduled_event
    
    
    async def scheduled_event_get_all_guild(self, guild):
        """
        Gets the given guild's scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild to get it's scheduled of.
        
        Returns
        -------
        scheduled_events : `list` of ``ScheduledEvent``
        
        Raises
        ------
        TypeError
            If `guild` is neither ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        scheduled_event_datas = await self.api.scheduled_event_get_all_guild(guild_id, {'with_user_count': True})
        return [ScheduledEvent.from_data(scheduled_event_data) for scheduled_event_data in scheduled_event_datas]
    
    
    async def scheduled_event_user_get_chunk(self, scheduled_event, *, after = None, before = None, limit = None):
        """
        Requests a chunk user subscribed to a scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        scheduled_event : ``ScheduledEvent | (int, int)``
            The scheduled event to get.
        after : ``None | int | DiscordEntity | DateTime`` = `None`, Optional (Keyword only)
            The timestamp after the subscribed users were created.
        before : ``None | int | DiscordEntity | DateTime`` = `None`, Optional (Keyword only)
            The timestamp before the subscribed users were created.
        limit : `None | int` = `None`, Optional (Keyword only)
            The amount of scheduled event users to request. Can be between 1 and 100.
        
        Returns
        -------
        users : ``list<ClientUserBase>``
        
        Raises
        ------
        TypeError
            - If `scheduled_event` is invalid type.
            - If `after`, `before` was passed with an unexpected type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `limit` was not given as `int`.
            - If `limit` is out of range [1:100].
        """
        guild_id, scheduled_event_id = get_scheduled_event_guild_id_and_id(scheduled_event)
        
        if limit is None:
            limit = 100
        else:
            if __debug__:
                if not isinstance(limit, int):
                    raise AssertionError(
                        f'`limit` can be `int`, got {limit.__class__.__name__}; {limit!r}.'
                    )
                
                if (limit < 1) or (limit > 100):
                    raise AssertionError(
                        f'`limit` is out from the expected [1:100] range, got {limit!r}.'
                    )
        
        query_parameters = {}
        
        if limit != 100:
            query_parameters['limit'] = limit
        
        if (after is not None):
            query_parameters['after'] = log_time_converter(after)
        
        if (before is not None):
            query_parameters['before'] = log_time_converter(before)
        
        if guild_id:
            query_parameters['with_member'] = True
        
        scheduled_event_user_datas = await self.api.scheduled_event_user_get_chunk(
            guild_id, scheduled_event_id, query_parameters
        )
        
        users = []
        for scheduled_event_user_data in scheduled_event_user_datas:
            user_data = scheduled_event_user_data['user']
            user = User.from_data(user_data, user_data.get('member', None), guild_id)
            users.append(user)
        
        return users
    
    
    async def scheduled_event_user_get_all(self, scheduled_event):
        """
        Requests all user subscribed to a scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        scheduled_event : ``ScheduledEvent | (int, int)``
            The scheduled event to get.
        
        Returns
        -------
        users : ``list<ClientUserBase>``
        
        Raises
        ------
        TypeError
            - If `scheduled_event` is invalid type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, scheduled_event_id = get_scheduled_event_guild_id_and_id(scheduled_event)
        
        query_parameters = {'after': 0}
        
        if guild_id:
            query_parameters['with_member'] = True
        
        users = []
        
        while True:
            scheduled_event_user_datas = await self.api.scheduled_event_user_get_chunk(
                guild_id, scheduled_event_id, query_parameters,
            )
            
            for scheduled_event_user_data in scheduled_event_user_datas:
                user_data = scheduled_event_user_data['user']
                user = User.from_data(user_data, user_data.get('member', None), guild_id)
                users.append(user)
            
            if len(scheduled_event_user_datas) < 100:
                break
            
            query_parameters['after'] = users[-1].id
        
        return users
    
    
    async def scheduled_event_occasion_overwrite_create(
        self,
        scheduled_event,
        timestamp,
        scheduled_event_occasion_overwrite_template = None,
        *,
        reason = None,
        **keyword_parameters,
    ):
        """
        Cancels a single occasion of a reoccurring scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        scheduled_event : ``ScheduledEvent | (int, int)``
            The scheduled event to cancel an occasion of.
        
        timestamp : `None | DateTime`
            The timestamp to create scheduled event overwrite for.
        
        scheduled_event_occasion_overwrite_template : ``None | ScheduledEventOccasionOverwrite``` = `None`, Optional
            Scheduled event occasion overwrite to use as a template.
        
        reason : `None | str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other Parameters
        ----------------
        cancelled : `None | bool``, Optional (Keyword only)
            Whether the occasion is cancelled.
        
        end : `None | DateTime` = `None`, Optional (Keyword only)
            When the event will end.
        
        start : `DateTime`, Optional (Keyword only)
            When the event will start.
        
        Raises
        ------
        TypeError
            - If `scheduled_event` is invalid type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, scheduled_event_id = get_scheduled_event_guild_id_and_id(scheduled_event)
        
        data = build_create_payload(
            scheduled_event_occasion_overwrite_template,
            SCHEDULED_EVENT_OCCASION_OVERWRITE_CREATE_FIELD_CONVERTERS,
            keyword_parameters,
        )
        
        data['original_scheduled_start_time'] = datetime_to_timestamp(validate_timestamp(timestamp))
        
        await self.api.scheduled_event_occasion_overwrite_create(
            guild_id,
            scheduled_event_id,
            data,
            reason,
        )
    
    
    async def scheduled_event_occasion_overwrite_edit(
        self,
        scheduled_event,
        timestamp,
        scheduled_event_occasion_overwrite_template = None,
        *,
        reason = None,
        **keyword_parameters,
    ):
        """
        Cancels a single occasion of a reoccurring scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        scheduled_event : ``ScheduledEvent | (int, int)``
            The scheduled event to cancel an occasion of.
        
        timestamp : `None | DateTime`
            The timestamp to edit.
        
        scheduled_event_occasion_overwrite_template : ``None | ScheduledEventOccasionOverwrite``` = `None`, Optional
            Scheduled event occasion overwrite to use as a template.
        
        reason : `None | str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other Parameters
        ----------------
        cancelled : `None | bool``, Optional (Keyword only)
            Whether the occasion is cancelled.
        
        end : `None | DateTime` = `None`, Optional (Keyword only)
            When the event will end.
        
        start : `DateTime`, Optional (Keyword only)
            When the event will start.
        
        Raises
        ------
        TypeError
            - If `scheduled_event` is invalid type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        scheduled_event, guild_id, scheduled_event_id = get_scheduled_event_and_guild_id_and_id(scheduled_event)
        if scheduled_event is None:
            scheduled_event_occasion_overwrite_original = None
        else:
            scheduled_event_occasion_overwrite_original = scheduled_event_occasion_overwrite_get(
                scheduled_event, timestamp
            )
        
        data = build_edit_payload(
            scheduled_event_occasion_overwrite_original,
            scheduled_event_occasion_overwrite_template,
            SCHEDULED_EVENT_OCCASION_OVERWRITE_EDIT_FIELD_CONVERTERS,
            keyword_parameters,
        )
        
        await self.api.scheduled_event_occasion_overwrite_edit(
            guild_id,
            scheduled_event_id,
            datetime_to_id(validate_timestamp(timestamp)),
            data,
            reason,
        )
    
    
    async def scheduled_event_occasion_overwrite_delete(self, scheduled_event, timestamp, *, reason = None):
        """
        Uncancels a single occasion of a reoccurring scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        scheduled_event : ``ScheduledEvent | (int, int)``
            The scheduled event to uncancel an occasion of.
        
        timestamp : `None | DateTime`
            The timestamp to delete.
        
        reason : `None | str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `scheduled_event` is invalid type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, scheduled_event_id = get_scheduled_event_guild_id_and_id(scheduled_event)
        
        await self.api.scheduled_event_occasion_overwrite_delete(
            guild_id,
            scheduled_event_id,
            datetime_to_id(validate_timestamp(timestamp)),
            reason,
        )



__all__ = ('InteractionComponentMetadataBase',)

from scarletio import RichAttributeErrorBaseType

from ...bases import PlaceHolder


class InteractionComponentMetadataBase(RichAttributeErrorBaseType):
    """
    Base type for interaction component metadata.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new interaction component metadata from the given fields.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return object.__new__(cls)
    
    
    @classmethod
    def from_keyword_parameters(cls, keyword_parameters):
        """
        Creates a new interaction component metadata from the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict<str, object>`
            Keyword parameters to build the metadata from.
        
        Returns
        -------
        self : `instance<type<cls>>`
        
        Raises
        ------
        TypeError
            - If a keyword parameter's type is incorrect.
        ValueError
            - If a keyword parameter's value is incorrect.
        """
        return object.__new__(cls)
    
    
    def __repr__(self):
        """Returns repr(self)."""
        return f'<{type(self).__name__}>'
    
    
    def __hash__(self):
        """Returns hash(self)."""
        return 0
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return False
        
        return self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self equals to other. Other must be the same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        return True
    
    
    def _match_to_component(self, other):
        """
        matches self to an other component.
        
        Parameters
        ----------
        other : ``ComponentMetadataBase``
        
        Returns
        -------
        matching : `bool`
        """
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new interaction component metadata from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Component data.
        
        Returns
        -------
        self : `instance<cls>`
            Returns the created component.
        """
        return object.__new__(cls)
    
    
    def to_data(self, *, defaults = False):
        """
        Returns the interaction component metadata's json serializable representation.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with default values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        return {}
    
    
    def copy(self):
        """
        Copies the interaction component metadata.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return object.__new__(type(self))
    
    
    def copy_with(self):
        """
        Copies the interaction component metadata with the given fields.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        return self.copy()
    
    
    def copy_with_keyword_parameters(self, keyword_parameters):
        """
        Copies the interaction component metadata with the given keyword parameters.
        
        The used up fields are popped from the keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict<str, object>`
            Keyword parameters defining which fields should be changed.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return self.copy_with()
    
    
    def iter_custom_ids_and_values(self):
        """
        Iterates over all the `custom_id`-s and values of the interaction component metadata.
        
        This method is an iterable generator.
        
        Yields
        ------
        item : `(str, ComponentType, None | str | tuple<str>)`
        """
        return
        yield
    
    
    component = PlaceHolder(
        None,
        """
        The sub-component nested inside. Applicable for the label component.
        
        Returns
        -------
        component : ``None | Component``
        """
    )
    
    
    components = PlaceHolder(
        None,
        """
        Sub-components nested inside. Applicable for the row component.
        
        Returns
        -------
        components : ``None | tuple<Component>``
        """
    )
    
    
    custom_id = PlaceHolder(
        None,
        """
        Custom identifier to detect which component was clicked (or used) by the user.
        
        > Mutually exclusive with the `sku_id` and `url` fields if the component is a buttons.
        
        Returns
        -------
        custom_id : `None | str`
        """
    )
    
    
    thumbnail = PlaceHolder(
        None,
        """
        The thumbnail or other accessory (button) of a section component.
        
        Returns
        -------
        thumbnail : ``None | Component``
        """
    )
    
    
    value = PlaceHolder(
        None,
        """
        The component's value defined by the user.
        
        Returns
        -------
        value : `None | str`
        """
    )
    
    
    values = PlaceHolder(
        None,
        """
        The component's values defined by the user.
        
        Returns
        -------
        values : `None | tuple<str>`
        """
    )



from .role import *
from .role_color_configuration import *
from .role_manager_metadata import *


__all__ = (
    *role.__all__,
    *role_color_configuration.__all__,
    *role_manager_metadata.__all__,
)



from .activity import *
from .allowed_mentions import *
from .application import *
from .application_command import *
from .audit_logs import *
from .auto_moderation import *
from .bases import *
from .builder import *
from .channel import *
from .client import *
from .component import *
from .embed import *
from .emoji import *
from .events import *
from .exceptions import *
from .gateway import *
from .guild import *
from .http import *
from .integration import *
from .interaction import *
from .invite import *
from .localization import *
from .message import *
from .oauth2 import *
from .onboarding import *
from .permission import *
from .poll import *
from .resolved import *
from .scheduled_event import *
from .soundboard import *
from .role import *
from .stage import *
from .sticker import *
from .user import *
from .voice import *
from .webhook import *

from .allowed_mentions import *
from .ansi_format import *
from .color import *
from .core import *
from .field_parsers import *
from .field_putters import *
from .field_validators import *
from .object_binding import *
from .payload_building import *
from .preconverters import *
from .precreate_helpers import *
from .utils import *


__all__ = (
    *activity.__all__,
    *allowed_mentions.__all__,
    *application.__all__,
    *application_command.__all__,
    *audit_logs.__all__,
    *auto_moderation.__all__,
    *bases.__all__,
    *builder.__all__,
    *channel.__all__,
    *client.__all__,
    *embed.__all__,
    *emoji.__all__,
    *component.__all__,
    *events.__all__,
    *exceptions.__all__,
    *gateway.__all__,
    *guild.__all__,
    *http.__all__,
    *integration.__all__,
    *interaction.__all__,
    *invite.__all__,
    *localization.__all__,
    *message.__all__,
    *oauth2.__all__,
    *onboarding.__all__,
    *permission.__all__,
    *poll.__all__,
    *resolved.__all__,
    *role.__all__,
    *scheduled_event.__all__,
    *soundboard.__all__,
    *stage.__all__,
    *sticker.__all__,
    *user.__all__,
    *voice.__all__,
    *webhook.__all__,
    
    *ansi_format.__all__,
    *core.__all__,
    *field_parsers.__all__,
    *field_putters.__all__,
    *field_validators.__all__,
    *color.__all__,
    *object_binding.__all__,
    *payload_building.__all__,
    *preconverters.__all__,
    *precreate_helpers.__all__,
    *utils.__all__,
)


# Deprecations

from ..utils.module_deprecation import deprecated_import

# 2025-04-25
deprecated_import(GuildBadge, 'UserClan')

del deprecated_import



__all__ = ()

from .compatibility import AUTO_RELOAD_SUPPORTED, INotifyFlags


AUTO_RELOAD_DELAY = 5.0


if AUTO_RELOAD_SUPPORTED:
    WATCH_MASK_CREATE = INotifyFlags.CREATE
    WATCH_MASK_DELETE = INotifyFlags.DELETE
    WATCH_MASK_DELETE_SELF = INotifyFlags.DELETE_SELF
    WATCH_MASK_MOVE_FROM = INotifyFlags.MOVED_FROM
    WATCH_MASK_MOVE_SELF = INotifyFlags.MOVE_SELF
    WATCH_MASK_MOVE_TO = INotifyFlags.MOVED_TO
    WATCH_MASK_UPDATE = INotifyFlags.MODIFY
    
else:
    WATCH_MASK_CREATE = 0
    WATCH_MASK_DELETE = 0
    WATCH_MASK_DELETE_SELF = 0
    WATCH_MASK_MOVE_FROM = 0
    WATCH_MASK_MOVE_SELF = 0
    WATCH_MASK_MOVE_TO = 0
    WATCH_MASK_UPDATE = 0


WATCH_MASK_GENERAL = (
    WATCH_MASK_CREATE | WATCH_MASK_DELETE | WATCH_MASK_DELETE_SELF | WATCH_MASK_MOVE_FROM | WATCH_MASK_MOVE_SELF |
    WATCH_MASK_MOVE_TO | WATCH_MASK_UPDATE
)



__all__ = ('SessionPlatformType', 'Status', )

from scarletio import copy_docs

from ...bases import Preinstance as P, PreinstancedBase


class Status(PreinstancedBase, value_type = str):
    """
    Represents a Discord user's status.
    
    Attributes
    ----------
    name : `str`
        The status's name.
    
    position : `int`
        Internal position of the status for sorting purposes.
    
    value : `str`
        The identifier value of the status.
    
    Type Attributes
    ---------------
    Each predefined status also can be accessed as a type attribute:
    
    +-----------------------+-----------+-----------+
    | Type attribute name   | position  | value     |
    +=======================+===========+===========+
    | online                | 0         | idle      |
    +-----------------------+-----------+-----------+
    | idle                  | 1         | idle      |
    +-----------------------+-----------+-----------+
    | dnd                   | 2         | dnd       |
    +-----------------------+-----------+-----------+
    | offline               | 3         | offline   |
    +-----------------------+-----------+-----------+
    | invisible             | 3         | invisible |
    +-----------------------+-----------+-----------+
    """
    __slots__ = ('position', )
    
    def __new__(cls, value, name = None, position = 4):
        """
        Creates a new status.
        
        Parameters
        ----------
        value : `str`
            The identifier value of the status.
        
        name : `None | str` = `None`, Optional
            The status's name.
        
        position : `int = `4`, Optional
            Internal position of the status for sorting purposes.
        """
        if name is None:
            name = value
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.position = position
        return self
    
    
    @copy_docs(PreinstancedBase.__lt__)
    def __lt__(self, other):
        if self is other:
            return False
        
        self_type = type(self)
        other_type = type(other)
        if self_type is other_type:
            pass
        
        elif issubclass(other_type, self_type.VALUE_TYPE):
            try:
                other = self_type.INSTANCES[other]
            except KeyError:
                return True
        
        else:
            return NotImplemented
        
        self_position = self.position
        other_position = other.position
        
        if self_position < other_position:
            return True
        
        if self_position > other_position:
            return False
        
        return (self.value < other.value)
    
    
    @copy_docs(PreinstancedBase.__gt__)
    def __gt__(self, other):
        if self is other:
            return False
        
        self_type = type(self)
        other_type = type(other)
        if self_type is other_type:
            pass
        
        elif issubclass(other_type, self_type.VALUE_TYPE):
            try:
                other = self_type.INSTANCES[other]
            except KeyError:
                return False
        
        else:
            return NotImplemented
        
        self_position = self.position
        other_position = other.position
        
        if self_position > other_position:
            return True
        
        if self_position < other_position:
            return False
        
        return (self.value > other.value)
    
    
    @copy_docs(PreinstancedBase.__eq__)
    def __eq__(self, other):
        if self is other:
            return True
        
        self_type = type(self)
        other_type = type(other)
        if self_type is other_type:
            pass
        
        elif issubclass(other_type, self_type.VALUE_TYPE):
            try:
                other = self_type.INSTANCES[other]
            except KeyError:
                return False
        
        else:
            return NotImplemented
    
        if self.position != other.position:
            return False
        
        if self.value != other.value:
            return False
        
        return True
    
    
    # predefined
    online = P('online', 'online', 0)
    idle = P('idle', 'idle', 1)
    dnd = P('dnd', 'dnd', 2)
    offline = P('offline', 'offline', 3)
    invisible = P('invisible', 'invisible', 3)


# Set default status
Status.INSTANCES[''] = Status.offline


class SessionPlatformType(PreinstancedBase, value_type = str):
    """
    Represents a session's platform type a user is logged in from.
    
    Attributes
    ----------
    name : `str`
        The name of the session platform type.
    
    value : `str`
        The Discord side identifier value of the session platform type.
    
    Type Attributes
    ---------------
    Every predefined session platform type can also be accessed as type attribute:
    
    +-----------------------+---------------+-----------+
    | Type attribute name   | name          | value     |
    +=======================+===============+===========+
    | none                  | none          |           |
    +-----------------------+---------------+-----------+
    | desktop               | desktop       | desktop   |
    +-----------------------+---------------+-----------+
    | embedded              | embedded      | embedded  |
    +-----------------------+---------------+-----------+
    | mobile                | mobile        | mobile    |
    +-----------------------+---------------+-----------+
    | web                   | web           | web       |
    +-----------------------+---------------+-----------+
    """
    __slots__ = ()

    # predefined
    none = P('', 'none')
    desktop = P('desktop', 'desktop')
    embedded = P('embedded', 'embedded')
    mobile = P('mobile', 'mobile')
    web = P('web', 'web')
